#!/usr/local/bin/python

import sys
import os
import re
import subprocess
from optparse import OptionParser

import ABAC
import Creddy

# Routines that support the various test scripts.  Mostly the details of
# interfacing to the ABAC and creddy libs, as well as come commonly used
# routines and classes.

# Get hold of the local creddy installation
def _find_exec(n, p):
    '''
    Find an executable (n) in path tuple (p).  The first executable is
    returned, or None if none is found.  Used at module initialization time
    to find creddy and openssl.
		    '''
    for d in p:
	path = "%s/%s" % (d, n)
	if os.access(path, os.X_OK):
	    return path
    else:
	return None

creddy = _find_exec('creddy', ('/usr/bin/', '/usr/local/bin'))


class simple_options(OptionParser):
    '''
    Simple option parser that parses --debug
    '''
    def __init__(self):
	OptionParser.__init__(self)
	self.add_option('--debug', action='store_true', dest='debug',
		default=False, help='Debugging output')

# Convenient for debugging creddy calls below
def join(words, sep=' '):
    return sep.join(words)

def init_principal_identities(principals=('CH', 'SA', 'AM', 'P'), debug=False):
    '''
    Create directories and identity certificates for the given principals
    '''
    cwd = os.getcwd()
    try:
	for p in principals:
	    if not os.access(p, os.F_OK):
		os.mkdir(p)
	    os.chdir(p)
	    if not os.access('%s_ID.pem' % p, os.F_OK):
		create_id(p, debug)
	    os.chdir('..')
    finally:
	os.chdir(cwd)

def _creddy_certs(p, subj):
    '''
    Convert the directory names of the principals into the file names of their
    certs and keys
    '''
    cert = Creddy.ID('%s/%s_ID.pem' % (p, p))
    cert.load_privkey('%s/%s_private.pem' % (p, p))
    subject = Creddy.ID('%s/%s_ID.pem' % (subj, subj))

    return cert,  subject

def create_id(name, debug=False):
    '''
    Use creddy to create an identity certificate.  This happens in the current
    directory.
    '''
    if debug:
	print >>sys.stderr, 'Creating ID cert for %s (stand by)' % name
    cert = Creddy.ID(name, 365 * 5)
    f = open(cert.cert_filename(), 'w')
    cert.write_cert(f)
    f.close()
    f = open(cert.privkey_filename(), 'w')
    cert.write_privkey(f)
    f.close()

def creddy_assign(p, role, subj, out, debug=False):
    '''
    Use creddy to create an attribute certificate for an assignemnt of
    attribute.
    '''
    if debug:
	print >>sys.stderr, 'Creating %s.%s <- %s' % (p, role, subj)

    cert, subject = _creddy_certs(p, subj)

    a = Creddy.Attribute(cert, role, 5 * 365)
    a.principal(subject.keyid())
    a.bake()
    f = open(out, 'w')
    a.write(f)
    f.close()

def creddy_role(p, right_role, subj, left_role, out, debug=False):
    '''
    Use creddy to create an attribute certificate for a simple deduction rule.
    '''
    if debug:
	print >>sys.stderr, 'Creating %s.%s <- %s.%s' % \
		(p, right_role, subj, left_role)

    cert, subject = _creddy_certs(p, subj)

    a = Creddy.Attribute(cert, right_role, 5 * 365)
    a.role(subject.keyid(), left_role)
    a.bake()
    f = open(out, 'w')
    a.write(f)
    f.close()

def creddy_linking_role(p, right_role, subj, linking_role, linked_role, 
	out, debug=False):
    '''
    Use creddy to create an attribute certificate for a linked deduction rule.
    '''
    if debug:
	print >>sys.stderr, 'Creating %s.%s <- (%s.%s).%s' % \
		(p, right_role, subj, linking_role, linked_role)

    cert, subject = _creddy_certs(p, subj)

    a = Creddy.Attribute(cert, right_role, 5 * 365)
    a.linking_role(subject.keyid(), linking_role, linked_role)
    a.bake()
    f = open(out, 'w')
    a.write(f)
    f.close()

class translator:
    '''
    A class that applies translations in it's self.trans member to input
    strings.
    '''
    def __init__(self):
	self.trans = ()

    def __call__(self, s):
	if not isinstance(s, basestring):
	    return s
	for r, sub in self.trans:
	    s = r.sub(sub, s)
	return s


class translate_cred(translator):
    '''
    Translate keyid to symbolic name.
    '''
    def __init__(self, d):
	self.trans = tuple([(re.compile(i), n) for n, i in d.items()])

class untranslate_cred(translator):
    '''
    Translate symbolic name to keyid .
    '''
    def __init__(self, d):
	# Sort table so that the longest names are translated into keyids
	# first.  This makes sure translations are longest match.
	table = [(n, i) for n, i in d.items()]
	table.sort(key=lambda x: len(x[0]), reverse=True)
	self.trans = tuple([(re.compile(n), i) for n, i in table])


def id_map(pdir):
    '''
    Use creddy to get the keyid associated with the principal whose credentials
    are stored in the given directory.
    '''
    cert = "%s/%s_ID.pem" % (pdir, pdir)
    if os.access(cert, os.F_OK):
	try:
	    return (Creddy.ID(cert).keyid(), pdir)
	except RuntimeError, e:
	    return (None, None)
    else:
	return (None, None)

def print_creds(creds, trans=None):
    '''
    Print a sequence of libabac credentials.  If trans is given, call it on
    each line of output, e.g. to convert keyids to names.
    '''
    for c in creds:
	out = '%s <- %s' % (c.head().string(), c.tail().string())
	if trans: out = trans(out)
	print out

def load_directory(context, cred_dir):
    '''
    Load a given directory into the ABAC context.  This is more forgiving of
    certificate names that do not fit strict convention than
    ABAC.Context.load_directory is.
    '''
    no_issuer = []
    # Load each credential using the loaders.  Note that credentials
    # imported before the identity certificate that signed them are
    # unprocessable.  We put them aside and retry after the rest of the
    # batch has been imported.
    for c in [ "%s/%s" % (cred_dir, f) for f in os.listdir(cred_dir)]:
	if context.load_id_file(c) != ABAC.ABAC_CERT_SUCCESS:
	    a_rv = context.load_attribute_file(c) 
	    if a_rv == ABAC.ABAC_CERT_MISSING_ISSUER:
		no_issuer.append(c)
    # Re-process the deferred credentials
    for c in no_issuer:
	if context.load_attribute_file(c) != ABAC.ABAC_CERT_SUCCESS:
	    print >>sys.stderr, \
		"Cannot load %s: Missing identity certificate?" % c

